@@ -12,8 +12,8 @@
 
 import json
 import re
+from unittest import mock
 
-import mock
 from testtools import matchers
 
 from keystoneauth1 import adapter
@@ -395,14 +395,14 @@ class DiscoverUtils(utils.TestCase):
 
     def test_version_between(self):
         def good(minver, maxver, cand):
-            self.assertTrue(discover._version_between(minver, maxver, cand))
+            self.assertTrue(discover.version_between(minver, maxver, cand))
 
         def bad(minver, maxver, cand):
-            self.assertFalse(discover._version_between(minver, maxver, cand))
+            self.assertFalse(discover.version_between(minver, maxver, cand))
 
         def exc(excls, minver, maxver, cand):
             self.assertRaises(excls,
-                              discover._version_between, minver, maxver, cand)
+                              discover.version_between, minver, maxver, cand)
 
         # candidate required
         exc(ValueError, (1, 0), (1, 0), None)
@@ -491,6 +491,40 @@ class VersionDataTests(utils.TestCase):
 
             self.assertTrue(mock.called_once)
 
+    def test_version_data_override_version_url(self):
+        # if the request url is versioned already, just return it.
+        self.requests_mock.get(
+            V3_URL, status_code=200,
+            json={'version': fixture.V3Discovery('http://override/identity/v3')
+                  }
+        )
+
+        disc = discover.Discover(self.session, V3_URL)
+        version_data = disc.version_data()
+
+        for v in version_data:
+            self.assertEqual(v['version'], (3, 0))
+            self.assertEqual(v['status'], discover.Status.CURRENT)
+            self.assertEqual(v['raw_status'], 'stable')
+            self.assertEqual(v['url'], V3_URL)
+
+        # if the request url is not versioned, just add version info to it.(
+        # do not changed the url's netloc or path)
+        self.requests_mock.get(
+            BASE_URL, status_code=200,
+            json={'version': fixture.V3Discovery('http://override/identity/v3')
+                  }
+        )
+
+        disc = discover.Discover(self.session, BASE_URL)
+        version_data = disc.version_data()
+
+        for v in version_data:
+            self.assertEqual(v['version'], (3, 0))
+            self.assertEqual(v['status'], discover.Status.CURRENT)
+            self.assertEqual(v['raw_status'], 'stable')
+            self.assertEqual(v['url'], V3_URL)
+
     def test_version_data_unknown(self):
         discovery_fixture = fixture.V3Discovery(V3_URL)
         discovery_fixture.status = 'hungry'
@@ -526,6 +560,102 @@ class VersionDataTests(utils.TestCase):
 
         self.assertTrue(mock.called_once)
 
+    def test_version_data_legacy_ironic_no_override(self):
+        """Validate detection of legacy Ironic microversion ranges."""
+        ironic_url = 'https://bare-metal.example.com/v1/'
+        self.requests_mock.get(
+            ironic_url, status_code=200,
+            json={
+                'id': 'v1',
+                'links': [{
+                    "href": ironic_url,
+                    "rel": "self"}]},
+            headers={
+                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
+                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
+            })
+
+        plugin = noauth.NoAuth()
+        a = adapter.Adapter(
+            self.session,
+            auth=plugin,
+            service_type='baremetal')
+
+        self.assertIsNone(a.get_api_major_version())
+
+    def test_version_data_ironic_microversions(self):
+        """Validate detection of Ironic microversion ranges."""
+        ironic_url = 'https://bare-metal.example.com/v1/'
+        self.requests_mock.get(
+            ironic_url, status_code=200,
+            json={
+                'id': 'v1',
+                'version': {
+                    'id': 'v1',
+                    'links': [{
+                        "href": ironic_url,
+                        "rel": "self"}],
+                    'version': '1.40',
+                    'min_version': '1.10',
+                    'status': 'CURRENT',
+                },
+                'links': [{
+                    "href": ironic_url,
+                    "rel": "self"}],
+            },
+            # Keep headers so we can verify that body trumps headers
+            headers={
+                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
+                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
+            })
+
+        self.assertEqual(
+            [
+                {
+                    'collection': None,
+                    'version': (1, 0),
+                    'url': ironic_url,
+                    'status': discover.Status.CURRENT,
+                    'raw_status': discover.Status.CURRENT,
+                    'min_microversion': (1, 10),
+                    'max_microversion': (1, 40),
+                    'next_min_version': None,
+                    'not_before': None,
+                },
+            ],
+            discover.Discover(self.session, ironic_url).version_data())
+
+    def test_version_data_legacy_ironic_microversions(self):
+        """Validate detection of legacy Ironic microversion ranges."""
+        ironic_url = 'https://bare-metal.example.com/v1/'
+        self.requests_mock.get(
+            ironic_url, status_code=200,
+            json={
+                'id': 'v1',
+                'links': [{
+                    "href": ironic_url,
+                    "rel": "self"}]},
+            headers={
+                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
+                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
+            })
+
+        self.assertEqual(
+            [
+                {
+                    'collection': None,
+                    'version': (1, 0),
+                    'url': ironic_url,
+                    'status': discover.Status.CURRENT,
+                    'raw_status': discover.Status.CURRENT,
+                    'min_microversion': (1, 3),
+                    'max_microversion': (1, 21),
+                    'next_min_version': None,
+                    'not_before': None,
+                },
+            ],
+            discover.Discover(self.session, ironic_url).version_data())
+
     def test_version_data_microversions(self):
         """Validate [min_|max_]version conversion to {min|max}_microversion."""
         def setup_mock(versions_in):
@@ -1111,6 +1241,29 @@ class EndpointDataTests(utils.TestCase):
             [mock.call('sess', url, cache='cache', authenticated=False)
              for url in ('url1', 'url2', 'url3')])
 
+    def test_run_discovery_auth(self):
+        url = 'https://example.com'
+        headers = {'Accept': 'application/json'}
+
+        session = mock.Mock()
+        session.get.side_effect = [
+            exceptions.Unauthorized('unauthorized'),
+            # Throw a different exception the second time so that we can
+            # catch it in the test and verify the retry.
+            exceptions.BadRequest('bad request'),
+        ]
+
+        try:
+            discover.get_version_data(session, url)
+        except exceptions.BadRequest:
+            pass
+        # Only one call with 'url'
+        self.assertEqual(2, session.get.call_count)
+        session.get.assert_has_calls([
+            mock.call(url, headers=headers, authenticated=None),
+            mock.call(url, headers=headers, authenticated=True),
+        ])
+
     def test_endpoint_data_str(self):
         """Validate EndpointData.__str__."""
         # Populate a few fields to make sure they come through.
@@ -1127,3 +1280,8 @@ class EndpointDataTests(utils.TestCase):
         self.assertEqual(exp, str(epd))
         # Works with implicit stringification
         self.assertEqual(exp, "%s" % epd)
+
+    def test_project_id_int_fallback(self):
+        bad_url = "https://compute.example.com/v2/123456"
+        epd = discover.EndpointData(catalog_url=bad_url)
+        self.assertEqual((2, 0), epd.api_version)
